/*
  (C)2019 izumi@ieee.org, all rights reserved
  CLIBS = -mwindows -lws2_32 -lmswsock -ladvapi32
*/

#define GEISTERCLIENT_C

#define DEBUG 0
#define SOCKLOG 1

#include <windows.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#include <unistd.h>
#include <time.h>
#include <stdio.h>
#include <string.h>

#include "geister.h"
#include "geisterclient.h"

#define BUFSZ 256

static WSADATA wsadata;
static SOCKET gsock=INVALID_SOCKET;

/* ---------------------------------------------------------------- */
int zeromemory(char *p, int l){
  int i;
  for (i=0;i<l;i++) p[i]=0;
  return TRUE;
}

/* ---------------------------------------------------------------- */
static FILE *socklog=NULL;

int GstSockLogOpen(void){
  time_t t;
  struct tm *d;
  char filename[256];

  if (!SOCKLOG) return FALSE;

  time(&t);
  d=localtime(&t);
  sprintf(filename,"socklog_%02d%02d%02d_%02d%02d%02d.txt",
	  d->tm_year-100,d->tm_mon+1,d->tm_mday,
	  d->tm_hour,d->tm_min,d->tm_sec);
  socklog=fopen(filename,"w");
  if (socklog==NULL) {
    fprintf(stderr,"unable to open socket log file (%s)\n",filename);
    return FALSE;
  }

  fprintf(socklog,"socket communication log\n20%02d/%02d/%02d %02d:%02d:%02d\n",
	  d->tm_year-100,d->tm_mon+1,d->tm_mday,
	  d->tm_hour,d->tm_min,d->tm_sec);

  return TRUE;
}

int GstSockLogClose(void){
  if (!SOCKLOG) return FALSE;
  fclose(socklog);
  socklog=NULL;
  return TRUE;
}

/* ---------------------------------------------------------------- */
int GstConnectHost(char *hostaddr, char *hostport){
  struct addrinfo hints, *arslt=NULL, *aptr=NULL;
  int r=0;

  if (DEBUG>0) fprintf(stderr,"GstConnectHost()\n");
  if (SOCKLOG) if (!GstSockLogOpen()) GstDie("unable to open socket log");
  if (SOCKLOG) fprintf(socklog,"connecting to %s:%s\n",hostaddr,hostport);

  gsock=INVALID_SOCKET;

  zeromemory((char*)&hints,sizeof(hints));
  hints.ai_family  =AF_UNSPEC;
  hints.ai_socktype=SOCK_STREAM;
  hints.ai_protocol=IPPROTO_TCP;

  // initialize winsock
  r=WSAStartup(MAKEWORD(2,2),&wsadata);
  if (r!=0) {
    fprintf(stderr,"WSAStartup failed with error: %d\n",r);
    return FALSE;
  }

  //resolve the server address and port
  r=getaddrinfo(hostaddr,hostport,&hints,&arslt);
  if (r!=0) {
    fprintf(stderr,"getaddrinfo failed with error: %d\n",r);
    WSACleanup();
    return FALSE;
  }

  //attempt to connect
  for(aptr=arslt; aptr!=NULL ;aptr=aptr->ai_next) {
    //create a SOCKET for connecting to server
    gsock=socket(aptr->ai_family,aptr->ai_socktype,aptr->ai_protocol);
    if (gsock==INVALID_SOCKET) {
      fprintf(stderr,"socket failed with error: %d\n", WSAGetLastError());
      WSACleanup();
      return FALSE;
    }
    //connect to server
    r=connect(gsock,aptr->ai_addr,(int)aptr->ai_addrlen);
    if (r!=SOCKET_ERROR) break; // connected successfully
    closesocket(gsock);
    gsock=INVALID_SOCKET;
  }
  freeaddrinfo(arslt);

  if (gsock==INVALID_SOCKET) {
    fprintf(stderr,"unable to connect to server!\n");
    WSACleanup();
    return FALSE;
  }

  fprintf(socklog,"socket connected successfully\n");

  return TRUE;
}


int GstCloseHost(void){
  if (DEBUG>0) fprintf(stderr,"GstCloseHost()\n");
  closesocket(gsock);
  gsock=INVALID_SOCKET;
  WSACleanup();
  if (SOCKLOG) fprintf(socklog,"closing socket\n");
  if (SOCKLOG) GstSockLogClose();

  return 0;
}

/* ---------------------------------------------------------------- */
int GstRecvLine(char buf[]){
  int r,i;

  if (DEBUG>64) fprintf(stderr,"GstRecvLine()\n");

  for (i=0;i<BUFSZ;i++) buf[i]='\0';

  r=recv(gsock,buf,BUFSZ-1,0);
  if (r==SOCKET_ERROR) GstDie("error in recv(socket)\n");
  buf[r]='\0';
  if (r==0) fprintf(stderr,"received empty packet\n");
  if (SOCKLOG) fprintf(socklog,"< %s",buf);

  return r;
}


int GstSendLine(char buf[]){
  int r;
  if (DEBUG>64) fprintf(stderr,"GstSendLine()\n");
  r=send(gsock,buf,strlen(buf),0);
  if (r==SOCKET_ERROR) GstDie("error in send(socket)\n");
  if (SOCKLOG) fprintf(socklog,"> %s",buf);

  return r;
}

/* ---------------------------------------------------------------- */
int GstSendGhostInfo(struct GstGameStr *s) {
  struct GstGhostStr *g;
  char buf[BUFSZ];
  int i,j;
  
  if (DEBUG>2) fprintf(stderr,"GstSendGhostInfo()\n");

  //receive message (SET?)
  GstRecvLine(buf);
  if (strncmp(buf,"SET?",4)!=0)
    GstDie("GstSendGhostInfo: expect (SET?) but recv wrong msg\n");
  
  // send ghost setting
  strncpy(buf,"SET:XXXX\r\n",BUFSZ);
  j=4;// first X
  for (i=0;i<8;i++) {
    g=&(s->ghost[GstPly0][i]);
    if (g->color==GstRed) buf[j++]='A'+g->id;
  }
  GstSendLine(buf);

  //receive message (OK?)
  GstRecvLine(buf);
  if (strncmp(buf,"OK",2)!=0)
    GstDie("GstSendGhostInfo: expect (OK) but recv wrong msg\n");
  
  return TRUE;
}

/* ---------------------------------------------------------------- */
int GstRecvBoardState(struct GstGameStr *s){
  struct GstGhostStr *g;
  char buf[BUFSZ];
  int result=GstContinue,ofs;
  int p,i,c,x,y;

  if (DEBUG>2) fprintf(stderr,"GstRecvBoardState()\n");

  //receive message 
  GstRecvLine(buf);
  if (strncmp(buf,"MOV",3)==0)      result=GstContinue;
  else if (strncmp(buf,"WON",3)==0) result=GstWin;
  else if (strncmp(buf,"LST",3)==0) result=GstLose;
  else GstDie("GstRecvBoardState: expect (MOV|WON|LST) but recv wrong msg\n");

  //clear board
  for (y=0;y<GstBdH;y++)
    for (x=0;x<GstBdW;x++)
      s->board[x][y]=NULL;

  //clear capture count
  for (p=0;p<2;p++)
    for (c=0;c<2;c++)
      s->capturecount[p][c]=0;

  //translate ghost info
  for (p=0;p<2;p++) {
    for (i=0;i<8;i++) {
      g=&(s->ghost[p][i]);
      g->player=p;
      g->id=i;
      ofs=4+p*24+i*3;
      switch (buf[ofs+2]){
      case 'R':case 'r':c=GstRed;    break;
      case 'B':case 'b':c=GstBlue;   break;
      case 'U':case 'u':c=GstUnknown;break;
      default: GstDie("GstRecvBoardState: wrong color\n");
      }
      g->color=c;
      x=buf[ofs+0]-'0';
      y=buf[ofs+1]-'0';
      if (x>=0&&x<=5&&y>=0&&y<=5) {
	// on board
	g->x=x;
	g->y=y;
	g->state=GstOnboard;
	s->board[x][y]=g;
      } else if (x==8&&y==8) {
	// escaped
	g->x=-1;
	g->y=-1;
	g->state=GstEscaped;
      } else if (x==9&&y==9) {
	// captured
	g->x=-1;
	g->y=-1;
	g->state=GstCaptured;
	if (c==GstUnknown) GstDie("GstRecvBoardState: captured but color unknown\n");
	s->capturecount[p][c]++;
      } else {
	GstDie("GstRecvBoardState: ghost is out of board");
      }
    }
  }

  s->turn+=1;

  return result;
}

/* ---------------------------------------------------------------- */
int GstSendAction(struct GstActionStr *a){
  char buf[BUFSZ];
  struct GstGhostStr *g;
  int d;

  if (DEBUG>2) fprintf(stderr,"GstSendAction()\n");

  g=a->ghost;
  d=a->dir;

  if (g->player!=GstPly0)
    GstDie("GstSendAction: wrong player\n");

  // send ghost setting
  strncpy(buf,"MOV:X,D\r\n",BUFSZ);
  buf[4]='A'+g->id;
  switch(d){
  case GstNorth:buf[6]='N';break;
  case GstEast :buf[6]='E';break;
  case GstSouth:buf[6]='S';break;
  case GstWest :buf[6]='W';break;
  default:GstDie("GstSendAction: wrong direction");
  }
  GstSendLine(buf);

  //receive message (OK?)
  GstRecvLine(buf);
  if (strncmp(buf,"OK",2)!=0)
    GstDie("GstSendAction: expect (OK) bug recv wrong msg\n");

  return TRUE;
}

/* ---------------------------------------------------------------- */
